"""
# Errors - Error handling for the supamodel package

This module contains custom error classes and functions for the supamodel package. These classes and functions are designed to provide more detailed and informative error messages, and to provide a consistent interface for handling errors across the package.

"""



import logging

import devtools
from loguru import logger
from rich import print
from rich.logging import RichHandler

httpx_logger = logging.getLogger("httpx")
httpx_logger.disabled = True

# logger.configure(
#     handlers=[
#         {
#             "sink": RichHandler(rich_tracebacks=True),
#             "level": "DEBUG",
#             "format": "{time} - {level} - {message}",
#         }
#     ]
# )

# logger.debug("This is a debug message")
# logger.info("This is an info message")
# logger.warning("This is a warning message")
# logger.error("This is an error message")
# logger.critical("This is a critical message")


__all__ = ["logger", "devtools", "print"]



"""
# Supabase API Support Models

This module contains the API support models for the supamodel package and beyond. For instance, the `ResponseModel` class is used to represent the response from an API request. 


The `SupaModel` class is used to represent a model that is stored in a Supabase database. We'll also have DecoratedRequestBuilders and SupaTypes in this module. They'll represent components that are used to interact with the Supabase API.


"""

import uuid
from typing import Any, Dict, Generic, List, Optional, TypeVar

from postgrest._sync.request_builder import SyncQueryRequestBuilder
from postgrest.base_request_builder import (
    APIResponse,
    BaseFilterRequestBuilder,
    BaseRPCRequestBuilder,
    BaseSelectRequestBuilder,
    QueryArgs,
    SingleAPIResponse,
    _ReturnT,
)
from pydantic import BaseModel, ConfigDict, Field, field_serializer

from supamodel._abc import Data, Dataset
from supamodel._client import supabase
from supamodel.utils import cached_classattr, tableize

RespT = TypeVar("RespT", bound=Data)


class ResponseModel(Data, Generic[RespT]):
    data: list[dict[str, Any]] | None = []
    count: int | None = None
    error: str | None = None

    def is_error(self):
        """
        Check if there is an error.

        Returns:
            bool: True if there is an error, False otherwise.
        """
        return self.error is not None

    def empty(self) -> bool:
        """
        Check if the data is empty.

        Returns:
            bool: True if the data is empty, False otherwise.
        """
        return self.data is None or len(self.data) == 0

    def is_success(self) -> bool:
        return self.error is None and self.data is not None

    def is_failure(self) -> bool:
        return bool(self.error) and self.empty()

    def dataset(self) -> Dataset[RespT]:
        return Dataset[RespT](data=self.data)


class SupaTable(Data):

    @classmethod
    def table_name(cls) -> str:
        return tableize(cls.__name__)

    @cached_classattr
    def table(cls) -> SyncQueryRequestBuilder:
        return supabase.table(cls.table_name())

    @cached_classattr
    def class_type(cls):
        """
        Returns the class type of an instance created from the model. It must be an instance so tha we don't get the parent class type.

        It would get ModelMetaClass instead of the current model normally instead.
        """
        instance = cls.model_construct({})
        return type(instance)

    @cached_classattr
    def resolved_type(cls):
        """
        Returns the resolved type of an instance created from the model. It must be an instance so that we don't get the parent class type.

        It would get ModelMetaClass instead of the current model normally instead.
        """
        instance = cls.model_construct({})
        return type(instance)

    @cached_classattr
    def dataset(cls):
        """
        Returns the dataset associated with the class.
        """
        return Dataset[cls.class_type]


class SupaRecord(Data):
    id: uuid.UUID | None = None

    @classmethod
    def table_name(cls) -> str:
        return tableize(cls.__name__)

    @cached_classattr
    def table(cls) -> SyncQueryRequestBuilder:
        return supabase.table(cls.table_name())

    @cached_classattr
    def class_type(cls):
        """
        Returns the class type of an instance created from the model. It must be an instance so tha we don't get the parent class type.

        It would get ModelMetaClass instead of the current model normally instead.
        """
        instance = cls.model_construct({})
        return type(instance)

    @cached_classattr
    def resolved_type(cls):
        """
        Returns the resolved type of an instance created from the model. It must be an instance so that we don't get the parent class type.

        It would get ModelMetaClass instead of the current model normally instead.
        """
        instance = cls.model_construct({})
        return type(instance)

    @cached_classattr
    def dataset(cls):
        """
        Returns the dataset associated with the class.
        """
        return Dataset[cls.class_type]



"""
# Abstract Base Classes

This module contains the abstract base classes for the supamodel package. These classes will help define the core functionality and structure of the models used in the package.

The primary abstract base classes defined in this module are:

* `BaseModel`: An abstract base class for Pydantic models that provides additional functionality and customization options.
* `OrmModel`: An abstract base class for models that interact with databases using an an existing ORM if necessary.
* `Data`: An abstract base class for data models that can handle dynamic fields and aliasing. This class is designed to be flexible and accommodate various data structures between systems.
* `Singleton`: An abstract base class for creating singleton classes that ensure only one instance of the class is created. We work to ensure that the class is thread-safe and can be used in a multi-threaded environment.
* `SupaModel`: An abstract base class for models that interact with Supabase using the Supabase API. This class provides a consistent interface for interacting with Supabase and handling errors.

These abstract base classes provide a foundation for the models used in the supamodel package, ensuring consistency, flexibility, and extensibility across different types of models.


"""

import random
from functools import cached_property
from typing import Any, Dict, Generic, List, Optional, TypeVar, cast

import pandas as pd
from pydantic import (
    AliasGenerator,
    BaseModel as _BaseModel,
    ConfigDict,
    EmailStr,
    alias_generators,
)

from supamodel.utils import cached_classattr

DataT = TypeVar("DataT", bound="Data")
InputModelT = TypeVar("InputModelT", "BaseModel", "Data")
DataAny = DataT | Any


class BaseModel(_BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        use_enum_values=True,
        extra="allow",
        populate_by_name=True,
        ignored_types=(cached_property, cached_classattr),
    )

    def none_dump(self) -> Dict:
        """Return the model's data in a dictionary format without incuding None values.

        Returns:
            Dict: A dict with the model's data without None values.
        """
        return self.model_dump(exclude_none=True)

    def supa_dump(self, by_alias: bool = False) -> Dict:
        """Supabase serialization dump.

        Args:
            by_alias (bool, optional): Option to use aliases. Defaults to False.

        Returns:
            Dict: The response in a dictionary format.
        """
        return self.model_dump(exclude_none=True, mode="json", by_alias=by_alias)


class IgnoreModel(BaseModel):
    model_config = ConfigDict(extra="ignore")


class CamelModel(IgnoreModel):
    model_config = ConfigDict(alias_generator=alias_generators.to_camel)


class OrmModel(BaseModel):
    model_config = ConfigDict(
        from_attributes=True,
        ignored_types=(cached_property, cached_classattr),
        arbitrary_types_allowed=True,
    )


class Data(OrmModel):
    """
    The OpenBB Standardized Data Model.

    The `Data` class is a flexible Pydantic model designed to accommodate various data structures
    for OpenBB's data processing pipeline as it's structured to support dynamic field definitions.

    The model leverages Pydantic's powerful validation features to ensure data integrity while
    providing the flexibility to handle extra fields that are not explicitly defined in the model's
    schema. This makes the `Data` class ideal for working with datasets that may have varying
    structures or come from heterogeneous sources.

    Key Features:
    - Dynamic field support: Can dynamically handle fields that are not pre-defined in the model,
        allowing for great flexibility in dealing with different data shapes.
    - Alias handling: Utilizes an aliasing mechanism to maintain compatibility with different naming
        conventions across various data formats.

    Usage:
    The `Data` class can be instantiated with keyword arguments corresponding to the fields of the
    expected data. It can also parse and validate data from JSON or other serializable formats, and
    convert them to a `Data` instance for easy manipulation and access.

    Example:
        # Direct instantiation
        data_record = Data(name="OpenBB", value=42)

        # Conversion from a dictionary
        data_dict = {"name": "OpenBB", "value": 42}
        data_record = Data(**data_dict)

    The class is highly extensible and can be subclassed to create more specific models tailored to
    particular datasets or domains, while still benefiting from the base functionality provided by the
    `Data` class.

    Attributes:
        __alias_dict__ (Dict[str, str]):
            A dictionary that maps field names to their aliases,
            facilitating the use of different naming conventions.
        model_config (ConfigDict):
            A configuration dictionary that defines the model's behavior,
            such as accepting extra fields, populating by name, and alias
            generation.
    """

    # __alias_dict__: Dict[str, str] = {}

    # def __repr__(self):
    #     """Return a string representation of the object."""
    #     return f"{self.__class__.__name__}({', '.join([f'{k}={v}' for k, v in super().model_dump().items()])})"

    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
        strict=False,
        alias_generator=AliasGenerator(
            validation_alias=alias_generators.to_camel,
            serialization_alias=alias_generators.to_snake,
        ),
    )


class Dataset(OrmModel, Generic[DataT]):
    # model_config = ConfigDict(
    #     from_attributes=True, arbitrary_types_allowed=True, ignored_types=[]
    # )
    # I'm now forcing the data to be a list of DataT
    # This is a breaking change. Luckily it's not used anywhere yet.
    data: Optional[List[DataT]] = []
    message: Optional[str] = None
    error: Optional[str] = None

    @property
    def is_list(self) -> bool:
        return isinstance(self.data, list)

    @property
    def is_dict(self) -> bool:
        return isinstance(self.data, dict)

    @property
    def is_base_model(self) -> bool:
        return isinstance(self.data, BaseModel)

    def sample(self) -> DataT:
        """
        Selects and returns a random item from the data.

        If the data is a list, a random item from the list is returned.
        If the data is a dictionary, a random item from the dictionary values is returned.
        If the data is neither a list nor a dictionary, the data itself is returned.

        Returns:
            Any: A random item from the data.
        """
        if self.is_list:
            return random.choice(self.data)
        return self.data

    # Add iterator methods if the data is a list
    def __iter__(self):
        if self.is_list:
            return iter(self.data)
        return iter([])

    def __getitem__(self, item):
        if self.is_list or self.is_dict:
            return self.data[item]
        return None

    def __len__(self):
        if self.is_list:
            return len(self.data)
        return 0

    def __contains__(self, item):
        if self.is_list:
            return item in self.data
        return False

    def empty(self) -> bool:
        if self.is_list:
            return len(self.data) == 0
        return True

    def to_dict(self) -> Dict[str, DataT] | List[DataT | Any] | Any:
        if self.is_list:
            self.data = cast(List[DataT], self.data)
            return [item.model_dump(by_alias=True) for item in self.data]
        elif self.is_dict:
            return self.data
        elif self.is_base_model:
            return self.data.model_dump(by_alias=True)
        return self.data

    def to_df(self) -> Any:
        if self.empty():
            return pd.DataFrame([])

        dataset_list = self.to_dict()
        if not isinstance(dataset_list, list):
            dataset_list = [dataset_list]
        return pd.DataFrame(dataset_list)


# -----------------------------------------------------------
# Local Testing
# -----------------------------------------------------------


class User(Data):
    username: str
    email: EmailStr
    age: int


def main():
    pass


if __name__ == "__main__":
    main()



import re
from datetime import datetime
from typing import Dict, Generic, List, Optional, TypeVar

import pandas as pd
import pendulum as pen
from pydantic import (
    BaseModel as _BaseModel,
    ConfigDict,
    Field,
    computed_field,
    field_validator,
)
from pydantic.alias_generators import to_camel as to_camel_case

from supamodel.enums import TimeInt


class BaseModel(_BaseModel):
    model_config: ConfigDict = ConfigDict(
        arbitrary_types_allowed=True,
        alias_generator=to_camel_case,
        populate_by_name=True,
        extra="ignore",
        use_enum_values=True,
    )


class Bar(BaseModel):
    address: str = Field(..., exclude=True)
    type: TimeInt
    close: float = Field(..., alias="c")
    high: float = Field(..., alias="h")
    low: float = Field(..., alias="l")
    open: float = Field(..., alias="o")
    volume: float = Field(..., alias="v")
    timestamp: datetime = Field(..., alias="unixTime")


class OhlcvData(BaseModel):
    items: List[Bar]


class OhlcvResponse(BaseModel):
    data: OhlcvData
    success: bool


class OverviewExtensions(BaseModel):
    coingecko_id: Optional[str] = None
    website: Optional[str] = None
    telegram: Optional[str] = None
    twitter: Optional[str] = None
    description: Optional[str] = None
    discord: Optional[str] = None
    medium: Optional[str] = None


class TokenItem(BaseModel):
    address: Optional[str] = None
    decimals: Optional[int] = None
    market_cap: Optional[float] = Field(None, alias="mc")
    symbol: Optional[str] = None
    name: Optional[str] = None
    last_trade: Optional[datetime] = Field(None, alias="lastTradeUnixTime")
    volume_change: Optional[float] = Field(0.0, alias="v24hChangePercent")


class TokenList(BaseModel):
    updateUnixTime: Optional[int] = 0
    updateTime: Optional[str] = 0
    tokens: Optional[List[TokenItem]] = 0
    total: Optional[int] = 0


# Method that checks if string is a valid URL
def is_url(s: str) -> bool:
    return s.startswith("http://") or s.startswith("https://")


class AssetOverview(BaseModel):
    address: Optional[str] = None
    decimals: Optional[int] = None
    symbol: Optional[str] = None
    name: Optional[str] = None
    logo_uri: Optional[str] = None
    liquidity: Optional[float] = None
    price: Optional[float] = None
    sell_amount: Optional[int] = None
    buy_amount: Optional[int] = None
    supply: Optional[float] = None
    market_cap: Optional[float] = None
    buy_volume: Optional[float] = None
    sell_volume: Optional[float] = None
    views: Optional[int] = None
    unique_views: Optional[float] = Field(None, allow_inf_nan=True)
    unique_wallet: Optional[float] = None
    volume: Optional[float] = None
    volume_usd: Optional[float] = None
    number_markets: Optional[int] = None
    price_change1h_percent: Optional[float] = None
    total_trades: Optional[int] = None
    total_trade_volume: Optional[float] = None

    model_config: ConfigDict = ConfigDict(extra="ignore")

    @field_validator("unique_views", mode="before")
    @classmethod
    def check_unique_views(cls, value):
        if (
            value is None
            or value == float("nan")
            or value == float("inf")
            or value == float("-inf")
        ):
            return 0
        return value


class Overview(BaseModel):
    """
    Represents overview data for a specific entity.

    Attributes:
        address (Optional[str]): The address of the entity.
        decimals (Optional[int]): The number of decimal places for the entity's value.
        symbol (Optional[str]): The symbol representing the entity.
        name (Optional[str]): The name of the entity.
        extensions (Optional[OverviewExtensions]): Additional extensions for the entity.
        logo_uri (Optional[str]): The URI of the entity's logo.
        liquidity (Optional[float]): The liquidity value of the entity.
        price (Optional[float]): The price of the entity.
        sell_amount (Optional[int]): The amount of entity sold in the last hour.
        buy_amount (Optional[int]): The amount of entity bought in the last hour.
        supply (Optional[float]): The supply of the entity.
        market_cap (Optional[float]): The market capitalization of the entity.
        buy_volume (Optional[float]): The volume of entity bought in the last hour.
        sell_volume (Optional[float]): The volume of entity sold in the last hour.
        watch (Optional[int]): The number of times the entity is being watched.
        views (Optional[int]): The number of views in the last hour.
        unique_views (Optional[int]): The number of unique views in the last hour.
        unique_wallet (Optional[int]): The number of unique wallets in the last hour.
        volume (Optional[float]): The volume of the entity in the last hour.
        volume_usd (Optional[float]): The volume of the entity in USD in the last hour.
        number_markets (Optional[int]): The number of markets the entity is listed in.
        price_change1h_percent (Optional[float]): The percentage change in price in the last hour.

    Properties:
        total_trades: The total number of trades made in the last hour.
        total_trade_volume: The total trade volume in the last hour.

    Methods:
        get_extensions: Returns the extensions of the entity.

    """

    address: Optional[str] = None
    decimals: Optional[int] = None
    symbol: Optional[str] = None
    name: Optional[str] = None
    extensions: Optional[OverviewExtensions] = Field(None, exclude=True)
    logo_uri: Optional[str] = Field(None, alias="logoURI")
    # pay attention to this number as well
    liquidity: Optional[float] = None
    price: Optional[float] = None
    sell_amount: Optional[int] = Field(0, alias="sell1h")
    buy_amount: Optional[int] = Field(0, alias="buy1h")
    supply: Optional[float] = 0.0
    market_cap: Optional[float] = Field(0.0, alias="mc")
    # Can get this volume over time and see what's rising
    # Buy volume might be wrong. Will have to look at normal volume to made decisions
    buy_volume: Optional[float] = Field(None, alias="vBuy1h")
    sell_volume: Optional[float] = Field(None, alias="vSell1h")
    views: Optional[int] = Field(None, alias="view1h")
    unique_views: Optional[int] = Field(None, alias="uniqueView1h")
    unique_wallet: Optional[int] = Field(
        None,
        alias="uniqueWallet1h",
    )
    volume: Optional[float] = Field(None, alias="v1h")
    volume_usd: Optional[float] = Field(None, alias="v1hUSD")
    # Check to see the distribution of the number of markets
    # Most are 1-2 markets, there are probably some that are a lot more
    number_markets: Optional[int] = None
    price_change1h_percent: Optional[float] = Field(None, alias="priceChange1hPercent")

    @computed_field
    @property
    def total_trades(self) -> int:
        """Get the total number of trades that are made in the last hour"""
        if not self.buy_amount or not self.sell_amount:
            return 0
        return self.buy_amount + self.sell_amount

    @computed_field
    @property
    def total_trade_volume(self) -> float:
        # Look for lots of this perhaps you can see a trend
        """Get the total trade volume in the last hour"""
        if not self.buy_volume or not self.sell_volume:
            return 0.0
        return self.buy_volume + self.sell_volume

    @computed_field
    @property
    def links(self) -> Dict[str, str]:
        extensions = self.get_extensions()
        if not extensions:
            return {}

        return {
            key: val
            for key, val in extensions.items()
            if val is not None and is_url(val)
        }

    def get_extensions(self):
        """
        Returns the extensions of the entity.

        Returns:
            Optional[dict]: The extensions of the entity, or None if there are no extensions.
        """
        if not self.extensions:
            return None
        return self.extensions.model_dump(exclude_none=True)


class OverviewResponse(BaseModel):
    data: Optional[Overview]
    success: Optional[bool]


class SecurityData(BaseModel):
    creation_time: Optional[datetime] = None
    creator_balance: Optional[float] = None

    owner_balance: Optional[float] = None
    # Probably want these to be about average
    owner_percentage: Optional[float] = None
    top10_user_percent: Optional[float] = None
    creator_percentage: Optional[float] = None

    freezeable: Optional[bool] = False
    pre_market_holder: Optional[List] = []
    lock_info: Optional[Dict] = {}
    non_transferable: Optional[bool] = None


class SecurityResponse(BaseModel):
    data: Optional[SecurityData] = None
    success: Optional[bool] = None
    status_code: Optional[int] = None


class TokenData(BaseModel):
    address: Optional[str] = None
    value: Optional[float] = None
    timestamp: Optional[datetime] = Field(None, alias="updateUnixTime")

    @computed_field
    @property
    def human_time(self) -> str:
        """Get the total trade volume in the last hour"""
        pen_time = pen.instance(self.timestamp)
        return pen_time.to_datetime_string()


class MultiPriceResponse(BaseModel):
    data: Optional[List[TokenData]] = None
    success: Optional[bool] = None


class PriceResponse(BaseModel):
    data: Optional[TokenData] = None
    success: Optional[bool] = None


class WalletToken(BaseModel):
    address: Optional[str] = None
    decimals: Optional[int] = None
    lamports: Optional[int] = Field(
        None, alias="balance", description="Balance in lamports. 1 billion = 1 SOL"
    )
    amount: Optional[float] = Field(None, alias="uiAmount", description="Amount in SOL")
    chain_id: Optional[str] = None
    name: Optional[str] = None
    symbol: Optional[str] = None
    logo_uri: Optional[str] = Field(None, alias="logoURI")
    price_usd: Optional[float] = None
    value_usd: Optional[float] = None


class WalletResponse(BaseModel):
    wallet_address: Optional[str] = Field(None, alias="wallet")
    balance: Optional[float] = Field(None, alias="totalUsd")
    items: Optional[List[WalletToken]] = []

    def is_time_df(self, time_df: pd.DataFrame) -> bool:
        columns = time_df.columns.tolist()
        if "timestamp" not in columns:
            return False

        timestamp_type = time_df["timestamp"].dtype
        if time_df["timestamp"].empty:
            return False

        # Check if the timestamp column is of type datetime or timestamp
        if timestamp_type != "datetime64[ns]":
            return False

        return True

    def items_df(self) -> pd.DataFrame:
        if not self.items:
            return pd.DataFrame([{}])
        _df = pd.DataFrame([item.model_dump() for item in self.items])
        columns = _df.columns.tolist()
        #
        if "timestamp" in columns:
            # Get the type of the timestamp column
            # timestamp_type = _df["timestamp"].dtype
            _df.set_index("timestamp", inplace=True)

        return _df


T = TypeVar("T", bound=BaseModel)


class ResponseContainer(BaseModel, Generic[T]):
    data: List[T] = []

    def df(self) -> pd.DataFrame:
        if not self.data:
            raise pd.DataFrame([{}])
        _df = pd.DataFrame([item.model_dump() for item in self.data])
        columns = _df.columns.tolist()
        #
        if "timestamp" in columns:
            # Get the type of the timestamp column
            # timestamp_type = _df["timestamp"].dtype
            _df.set_index("timestamp", inplace=True)

        return _df

    @computed_field
    def count(self) -> int:
        return len(self.data)


def main():
    pass


if __name__ == "__main__":
    main()



# """Get it SupaModel, because it's like a SuperModel, but Supa!"""

# import contextlib
# import uuid
# from functools import cached_property
# from typing import Any, Optional, Type, TypeVar

# from loguru import logger
# from postgrest._sync.request_builder import SyncQueryRequestBuilder
# from pydantic import ConfigDict, field_serializer, model_validator

# # from pydantic.dataclasses import dataclass
# from rich import print

# from galapy.app.utils.formatting import tableize
# from galapy.config import supa_client as supabase
# from galapy.models.abstracts.data import Data
# from galapy.models.abstracts.dataset import Dataset
# from galapy.utils.decorators import cached_classproperty
# from galapy.utils.fn_tools import in_all_dicts, serialize_dict

# # from galapy.models.supas.core import Exchange

# SupaModelT = TypeVar("SupaModelT", bound=Data)
# ReqBuilderT = TypeVar("ReqBuilderT", bound=SyncQueryRequestBuilder)
# # PartT = TypeVar("PartT")
# SupaItem = TypeVar("SupaItem")


# class ResponseModel(Data):
#     model_config = ConfigDict(from_attributes=True)
#     data: list[dict[str, Any]] | None = []
#     count: int | None = None
#     error: str | None = None

#     def is_error(self):
#         return self.error is not None

#     def is_success(self) -> bool:
#         return self.error is None and self.data is not None

#     def empty(self) -> bool:
#         return self.data is None or len(self.data) == 0


# class SupaModel(Data):

#     id: uuid.UUID | None = None

#     # Like Super Model, but Supa
#     @field_serializer("id")
#     def serialize_courses_in_order(exchange_id: uuid.UUID | None):
#         if exchange_id is None:
#             return None
#         return str(exchange_id)

#     @classmethod
#     def table_name(cls) -> str:
#         return tableize(cls.__name__)

#     @cached_classproperty
#     def table(cls) -> SyncQueryRequestBuilder:
#         return supabase.table(cls.table_name())

#     @cached_classproperty
#     def class_type(cls):
#         """
#         Returns the class type of an instance created from the model. It must be an instance so tha we don't get the parent class type.
#         It would get ModelMetaClass instead of the current model normally instead.
#         """
#         instance = cls.model_construct({})
#         return type(instance)

#     @cached_classproperty
#     def dataset(cls):
#         """
#         Returns the dataset associated with the class.
#         """
#         return Dataset[cls.class_type]

#     @classmethod
#     def sync_by_id(cls, id: str) -> SupaModelT:
#         return cls.find_one(query_params={"id": id})

#     @classmethod
#     def rpc(cls, method: str, data: dict[str, Any]) -> ReqBuilderT:
#         query: ReqBuilderT = supabase.rpc(method, params=data)
#         # self.execute_dataset(query)
#         return query
#         # return self.execute_dataset(query)

#     @classmethod
#     def select(cls, columns: list[str] = ["*"]) -> ReqBuilderT:
#         return cls.table.select(*columns)

#     @classmethod
#     def select_eq(
#         cls, columns: list[str] = ["*"], inputs: dict[str, Any] = {}
#     ) -> ReqBuilderT:
#         # Looks for exact matches in the table
#         builder = cls.select(columns)
#         if not inputs:
#             return builder
#         for key, value in inputs.items():
#             builder = builder.eq(key, value)
#         return builder

#     @classmethod
#     def extract(
#         cls, query: ReqBuilderT, type_var: Optional[Type[SupaModelT]] = None
#     ) -> Dataset[SupaModelT]:
#         # Since most supabase queries return an object with a data attribute, and count attribute, this method will parse that information and the containing model's data attribute (subclasses of SupaModel).
#         response = query.execute()
#         response = ResponseModel.model_validate(response)
#         if response.is_error():
#             raise Exception(response.error)
#         resp_data = response.data
#         if type_var is not None:
#             return Dataset[type_var](data=resp_data)
#         return cls.dataset(data=resp_data)

#     def sync(self) -> SupaModelT:
#         dump = self.model_dump(exclude_none=True, round_trip=True)
#         found = self.find_one(query_params=dump)

#         if found is None:
#             return None
#         # for key, value in found:
#         #     print((key, value))
#         for key, value in found:
#             with contextlib.suppress(AttributeError):
#                 setattr(self, key, value)
#         return found
#         # logger.success(found)

#     def save_current(self, record: dict = {}) -> Dataset[SupaModelT]:
#         try:
#             dump = (
#                 record if record else self.model_dump(exclude_none=True, by_alias=True)
#             )
#             saved = self.save_one(dump)
#             if saved.data:
#                 self.id = saved[0].id
#             else:
#                 return saved
#         except Exception:
#             raise

#     @classmethod
#     def save_one(cls, record: dict = {}) -> Dataset[SupaModelT]:
#         if not record:
#             raise ValueError(
#                 "No data to save. Please provide a record with something in the dictionary."
#             )
#         dump_id = record.pop("id", None)
#         if dump_id:
#             query_fn = cls.table.upsert
#             record = {**record, "id": str(dump_id)}
#         else:
#             query_fn = cls.table.insert

#         serialize_dict(record)
#         query = query_fn(record)
#         # logger.debug(query)
#         # logger.warning(cls.table_name())
#         return cls.extract(query)

#     @classmethod
#     def save_many(cls, records: list[dict] = []) -> Dataset[SupaModelT]:
#         if not records:
#             raise ValueError(
#                 "No data to save. Please provide a record with something in the dictionary."
#             )
#         if not isinstance(records, list) or not all(
#             isinstance(record, dict) for record in records
#         ):
#             raise ValueError("Please provide a list of records to save.")
#         all_ids = in_all_dicts(records, "id")
#         query_fn = cls.table.insert if not all_ids else cls.table().upsert

#         query = query_fn(records)
#         return cls.extract(query)

#     def delete(self, id: str | None = None) -> Dataset[SupaModelT]:
#         try:
#             delete_id = id if id is not None else self.id
#             return self.delete_one(delete_id)
#         except Exception as e:
#             raise e

#     @classmethod
#     def delete_one(cls, id: str | None = None) -> SupaModelT:

#         query = cls.table.delete().eq("id", id)
#         return cls.extract(query)

#     # Delete many records by id
#     @classmethod
#     def delete_many(cls, ids: list[str] = []) -> Dataset[SupaModelT]:
#         if not ids or not isinstance(ids, list):
#             raise ValueError("No ids to delete. Please provide a list of ids.")
#         query = cls.table.delete().in_("id", ids)
#         return cls.extract(query)

#     @classmethod
#     def list_all(cls) -> Dataset[SupaModelT]:

#         with contextlib.suppress(Exception):
#             builder = cls.table.select("*")
#             return cls.extract(builder)
#         return cls.dataset(data=[])

#     @classmethod
#     def list_paginated(cls, page: int = 0, per_page: int = 50) -> Dataset[SupaModelT]:
#         with contextlib.suppress(Exception):
#             response = cls.table.select("*").range(page, per_page).execute()
#             return cls.dataset(data=response.data)
#         return cls.dataset(data=[])
#         # except Exception as e:
#         #     raise e

#     @model_validator(mode="before")
#     @classmethod
#     def validate_model(cls, values: Any) -> Any:
#         return values

#     @classmethod
#     def find_one(cls, query_params: dict = {}, ordering=True) -> SupaModelT | None:

#         query = cls.select_eq(inputs=query_params)
#         query = query.limit(1)
#         dataset = cls.extract(query)
#         if not dataset.empty():
#             return dataset[0]
#         return None


# def main():
#     print("Hello, World!")


# if __name__ == "__main__":
#     main()



from typing import Optional
from uuid import UUID, uuid4

from pendulum import DateTime
from pydantic import Field

from supamodel._abc import BaseModel


class Portfolio(BaseModel):
    id: UUID = Field(
        default_factory=uuid4, description="Unique identifier of the portfolio"
    )
    user_id: Optional[UUID] = Field(
        description="ID of the user associated with the portfolio"
    )
    name: str = Field(description="Name of the portfolio")
    exchange_id: Optional[UUID] = Field(
        description="ID of the exchange associated with the portfolio"
    )
    base_currency_id: Optional[UUID] = Field(
        description="ID of the base currency of the portfolio"
    )
    created_at: DateTime | None = Field(
        default_factory=DateTime.now,
        description="Timestamp of when the portfolio was created",
    )
    updated_at: DateTime | None = Field(
        default_factory=DateTime.now,
        description="Timestamp of when the portfolio was last updated",
    )
    balance: float = Field(
        default=0.0, description="The total balance of the portfolio"
    )


class Position(BaseModel):
    id: Optional[UUID] = Field(description="Unique identifier of the position")
    portfolio_id: UUID = Field(description="ID of the associated portfolio")
    asset_id: UUID = Field(description="ID of the associated asset")
    quantity: float = Field(
        description="Quantity of the asset held in the position", alias="amount"
    )
    average_price: float = Field(
        alias="value", description="Average price of the asset in the position"
    )
    created_at: Optional[DateTime] = Field(
        description="Timestamp of when the position was created"
    )
    updated_at: Optional[DateTime] = Field(
        description="Timestamp of when the position was last updated"
    )



import uuid
from abc import ABC, abstractmethod
from datetime import datetime
from uuid import UUID

import pendulum
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    model_validator,
)

from supamodel._logging import logger


class Sizing(BaseModel):
    months: int = 0
    weeks: int = 0
    days: int = 0
    hours: int = 0
    minutes: int = 0
    seconds: int = 0


class Window(BaseModel):
    tail: datetime
    head: datetime


class Clock(BaseModel, ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    window_size: Sizing = Sizing(weeks=1)

    @abstractmethod
    def now(self):
        pass

    def now_local(self):
        return self.now("local")

    @computed_field
    @property
    def head(self) -> pendulum.datetime:
        return self.now()

    @computed_field
    @property
    def tail(self) -> pendulum.datetime:
        return self.now().subtract(**self.window_size.model_dump())

    @computed_field
    @property
    def sliding_window(self) -> Window:
        return Window(tail=self.tail, head=self.head)


class LiveClock(Clock):
    def now(self):
        return pendulum.now("utc")


class TestClock(Clock):
    initial_time: datetime = pendulum.now("utc")
    current_time: datetime = pendulum.now("utc")
    step_size: Sizing = Sizing(days=1)

    @model_validator(mode="after")
    def reset_time(self):
        if self.initial_time < self.current_time:
            self.current_time = self.initial_time

    def now(self):
        return pendulum.instance(self.current_time)

    def step(self):
        self.current_time = self.current_time.add(**self.step_size.model_dump())

    def reset(self):
        self.current_time = self.initial_time


class BacktestConfig(BaseModel):
    backtest_id: UUID = Field(default_factory=uuid.uuid4)
    start_time: datetime = pendulum.now("utc")
    end_time: datetime = pendulum.now("utc")

    model_config = ConfigDict(arbitrary_types_allowed=True)


def main_test_clock():
    """
    This function is the main entry point for testing the clock functionality.

    It initializes a `LiveClock` object, sets the current time and start time for a `BacktestConfig` object,
    and creates a `TestClock` object with specific window and step sizes. It then resets the test clock and
    performs a series of operations within a while loop.

    Note: The while loop is currently commented out in the code.

    We're just running out of time for today.


    Args:
        None

    Returns:
        None
    """
    live = LiveClock()
    logger.success(f"{live.tail.to_cookie_string()} - {live.head.to_cookie_string()}")
    current_time = pendulum.now("utc")
    start_time = current_time.subtract(years=1)
    backtest = BacktestConfig(start_time=start_time, end_time=current_time)

    test = TestClock(
        initial_time=backtest.start_time,
        window_size=Sizing(days=3, weeks=0),
        step_size=Sizing(days=1),
    )
    test.reset()
    # bar_dataset = OHLCVDataset()

    # while True:
    #     window = test.sliding_window
    #     tail_cookie = window.tail.to_cookie_string()
    #     head_cookie = window.head.to_cookie_string()
    #     if test.sliding_window.head >= backtest.end_time:
    #         logger.info("Backtest complete!")
    #         break
    #     result = bar_dataset.between(window.tail, window.head, "1 day")

    #     logger.success(
    #         f"Queries a resampled OHLCV between the window tail ('{tail_cookie}') and head ('{head_cookie}')."
    #     )
    #     print(result)
    #     test.step()
    #     print("\n\n")


def main():
    main_test_clock()


if __name__ == "__main__":
    main()
